/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: AbstractHost.java 325 2011-09-15 11:57:02Z chou $
 *
 */

package org.ow2.sirocco.vmm.agent.domain;

import java.util.Date;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.management.Notification;
import javax.management.ObjectName;

import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.agent.main.VirtManagerAgent;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.InvalidVMConfigException;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.VMConfigSpec;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VirtualCdromSpec;
import org.ow2.sirocco.vmm.api.VirtualDiskSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineConfigSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean;
import org.ow2.sirocco.vmm.api.monitoring.PerfMetric;
import org.ow2.sirocco.vmm.api.monitoring.PerfMetricInfo;

public abstract class AbstractHost extends ManagedResource implements HostMXBean {
    protected List<VMConfigSpec> virtualMachinesCreationRequests = new CopyOnWriteArrayList<VMConfigSpec>();

    protected AbstractHost(final ObjectName objectName) {
        super(objectName);
    }

    protected boolean vmCreationInProgress(final String name) {
        for (VMConfigSpec spec : this.virtualMachinesCreationRequests) {
            if (spec.getName().equals(name)) {
                return true;
            }
        }
        return false;
    }

    public long getReservedMemoryMB() {
        long result = 0;
        try {
            for (VirtualMachineMXBean vm : this.getResidentVMs()) {
                result += vm.getMemorySizeMB();
            }
        } catch (VMMException ex) {
        }
        for (VMConfigSpec vmSpec : this.virtualMachinesCreationRequests) {
            result += vmSpec.getMemorySizeMB();
        }
        return result;
    }

    abstract public void release();

    public void delete() {
        try {
            AgentCommon.getMBeanServer().unregisterMBean(this.objectName);
        } catch (Exception ex) {
            ManagedResource.logger.error(ex);
        }
        this.release();
        this.emitNotification(NotificationTypes.HOST_INVENTORY_DEL, "host delete", this.objectName);
        VirtManagerAgent.getInstance().setResourceTreeDirty();
    }

    protected void checkVMConfigSpec(final VirtualMachineConfigSpec vmConfigSpec) throws InvalidVMConfigException {
        if (vmConfigSpec.getName() == null) {
            throw new InvalidVMConfigException("VM name not specified");
        }
        if (vmConfigSpec.getDiskSpecs() == null) {
            throw new InvalidVMConfigException("No disk device specified");
        }
        for (VirtualDiskSpec diskSpec : vmConfigSpec.getDiskSpecs()) {
            // TODO
        }
        for (VirtualCdromSpec cdromSpec : vmConfigSpec.getCdromSpecs()) {
            // TODO
        }

    }

    public PerfMetricInfo[] listPerfMetricInfos() throws VMMException {
        ServerPool sp = (ServerPool) this.getServerPool();
        if (sp.getMonitoringDriver() == null) {
            org.ow2.sirocco.vmm.agent.domain.ManagedResource.logger.error("Monitoring service missing");
            return null;
        }
        return sp.getMonitoringDriver().listPerfMetrics(this);
    }

    public PerfMetric getPerfMetric(final String info) throws VMMException {
        ServerPool sp = (ServerPool) this.getServerPool();
        if (sp.getMonitoringDriver() == null) {
            org.ow2.sirocco.vmm.agent.domain.ManagedResource.logger.error("Monitoring service missing");
            return null;
        }
        return sp.getMonitoringDriver().getPerfMetric(this, info);

    }

    public PerfMetric[] getPerfMetrics(final String metricName, final Date startTime, final Date endTime, final long period,
        final ConsolidationFunction conFun) throws VMMException {
        ServerPool sp = (ServerPool) this.getServerPool();
        if (sp.getMonitoringDriver() == null) {
            org.ow2.sirocco.vmm.agent.domain.ManagedResource.logger.error("Monitoring service missing");
            return null;
        }
        return sp.getMonitoringDriver().getPerfMetrics(this, metricName, startTime, endTime, period, conFun.toString());

    }

    @Override
    synchronized public void emitNotification(final String type, final String message, final Object userData,
        final long timeStamp) {
        if (!type.equals(NotificationTypes.PERF_REPORT)) {
            ManagedResource.logger.debug("Emitting notification [\n\ttype=" + type + "\n\tmsg=" + message + "\n\tsource="
                + this.objectName + "\n\tuserdata=" + userData + "]");
        }

        // Monitoring
        if (type.equals(NotificationTypes.VM_ADD)) {
            ObjectName ob = (ObjectName) userData;
            ServerPool sp = (ServerPool) this.getServerPool();
            if (sp.getMonitoringDriver() != null) {
                sp.getMonitoringDriver().startVM(
                    ob.getKeyProperty("name").substring(ob.getKeyProperty("name").lastIndexOf('/') + 1,
                        ob.getKeyProperty("name").length()));
                // System.out.println(NotificationTypes.VM_ADD + ":" +
                // ob.getKeyProperty("name"));
            }
        }
        if (type.equals(NotificationTypes.VM_DEL)) {
            ObjectName ob = (ObjectName) userData;
            ServerPool sp = (ServerPool) this.getServerPool();
            if (sp.getMonitoringDriver() != null) {
                sp.getMonitoringDriver().stopVM(
                    ob.getKeyProperty("name").substring(ob.getKeyProperty("name").lastIndexOf('/') + 1,
                        ob.getKeyProperty("name").length()));
                // System.out.println(NotificationTypes.VM_DEL + ":" +
                // ob.getKeyProperty("name"));
            }
        }

        // Fin Monitoring

        Notification notification = new Notification(type, this.objectName, super.notificationSequence++, message);
        notification.setUserData(userData);
        notification.setTimeStamp(timeStamp);
        this.sendNotification(notification);
    }

}
