/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: DomainBuilder.java 362 2011-09-30 13:26:20Z chou $
 *
 */

package org.ow2.sirocco.vmm.agent.domain;

import java.io.File;
import java.io.FileInputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import javax.management.ObjectName;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.driver.Driver;
import org.ow2.sirocco.vmm.agent.driver.DriverFactory;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.agent.monitoring.api.MonitoringDriverAPI;
import org.ow2.sirocco.vmm.agent.monitoring.driver.MonitoringDriver;
import org.ow2.sirocco.vmm.agent.monitoring.driver.MonitoringDriverFactory;
import org.ow2.sirocco.vmm.api.VMMException;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * The class DomainBuilder constructs Domain objects from XML specifications.
 * TODO check the syntax of the XML file with a schema
 */
public final class DomainBuilder {
    static Logger logger = Logger.getLogger(DomainBuilder.class);

    static class DomainXMLHandler extends DefaultHandler {
        Stack<Domain> domainStack = new Stack<Domain>();

        Stack<ResourcePartition> resourcePartitionStack = new Stack<ResourcePartition>();

        Domain rootDomain, currentDomain;

        ServerPool currentServerPool;

        MonitoringDriverAPI currentMonitorable;

        ResourcePartition currentResourcePartition;

        StringBuffer currentPath = new StringBuffer();

        @Override
        public void startElement(final String uri, final String localName, final String qName, final Attributes attrs) {
            if (qName.equals("domain")) {
                String name = attrs.getValue("name");
                if (name == null) {
                    DomainBuilder.logger.error("Missing domain name, skipping domain definition");
                    this.currentDomain = null;
                    return;
                }
                this.currentPath.append("/" + name);

                ObjectName domainObjectName = MBeanObjectNamer.makeDomainName(this.currentPath.toString());
                Domain domain = new Domain(name, domainObjectName, DomainBuilder.attributes2Map(attrs));
                try {
                    AgentCommon.getMBeanServer().registerMBean(domain, domainObjectName);
                } catch (Exception ex) {
                    DomainBuilder.logger.error("Skipping domain " + name, ex);
                    this.currentDomain = null;
                    return;
                }
                DomainBuilder.logger.info("Added Domain " + domainObjectName);
                if (this.domainStack.isEmpty()) {
                    this.rootDomain = domain;
                } else {
                    this.domainStack.peek().addSubDomain(domain);
                }
                this.domainStack.push(domain);
                this.currentDomain = domain;
            } else if (qName.equals("serverPool")) {
                if (this.currentDomain != null) {
                    String name = attrs.getValue("name");
                    String driverName = attrs.getValue("driver");

                    if (name == null) {
                        DomainBuilder.logger.error("skipped server pool because of missing name");
                        return;
                    }
                    if (driverName == null) {
                        DomainBuilder.logger.error("skipped server pool because of missing driver name");
                        return;
                    }

                    ObjectName serverPoolObjectName = null;
                    serverPoolObjectName = MBeanObjectNamer.makeServerPoolName(this.currentPath.toString() + "/" + name);
                    Driver driver = DriverFactory.getInstance().newDriver(driverName);

                    if (driver != null) {
                        try {
                            this.currentServerPool = driver.newServerPool(name, serverPoolObjectName,
                                DomainBuilder.attributes2Map(attrs));

                        } catch (IllegalArgumentException ex) {
                            DomainBuilder.logger.error("Skipping ServerPool " + name + ":" + ex.getMessage());
                            this.currentServerPool = null;
                            return;
                        } catch (VMMException ex) {
                            DomainBuilder.logger.error("Skipping ServerPool " + name + ":" + ex.getMessage());
                            this.currentServerPool = null;
                            return;
                        }
                        try {
                            AgentCommon.getMBeanServer().registerMBean(this.currentServerPool, serverPoolObjectName);
                        } catch (Exception ex) {
                            DomainBuilder.logger.error("Skipping ServerPool " + name, ex);
                            this.currentServerPool = null;
                            return;
                        }
                        DomainBuilder.logger.info("Added ServerPool " + serverPoolObjectName);
                        this.currentDomain.addServerPool(this.currentServerPool);
                    } else {
                        DomainBuilder.logger.error("Driver " + driverName + " not found");
                        this.currentServerPool = null;
                    }
                    if (this.currentServerPool != null) {
                        this.resourcePartitionStack.push(this.currentServerPool);
                        this.currentResourcePartition = this.currentServerPool;
                    }
                }
            } else if (qName.equals("monitoring")) {
                String monitoringdriverName = attrs.getValue("driver");
                if (monitoringdriverName == null) {
                    DomainBuilder.logger.info("No monitoring for this server pool " + this.currentServerPool.getName());
                } else {
                    String args[] = {attrs.getValue("monitoringaddr"), attrs.getValue("monitoringRRD")};

                    MonitoringDriver mdriver = MonitoringDriverFactory.getInstance().newDriver(monitoringdriverName);

                    if (mdriver != null) {
                        try {
                            this.currentMonitorable = mdriver.newMonitoringService(monitoringdriverName);
                        } catch (IllegalArgumentException ex) {
                            DomainBuilder.logger.error("Skipping MonitoringDriver " + monitoringdriverName + ":"
                                + ex.getMessage());
                            this.currentServerPool = null;
                            return;
                        } catch (VMMException ex) {
                            DomainBuilder.logger.error("Skipping MonitoringDriver " + monitoringdriverName + ":"
                                + ex.getMessage());
                            this.currentServerPool = null;
                            return;
                        }

                        this.currentMonitorable.start(args);
                        DomainBuilder.logger.info("Added Monitoring service " + monitoringdriverName);
                        // For each serverPool, a monitoring service
                        // will be launch
                        this.currentServerPool.setMonitoringDriver(this.currentMonitorable);
                    } else {
                        DomainBuilder.logger.error("MonitoringDriver " + monitoringdriverName + " not found");
                        this.currentMonitorable = null;
                    }

                }

            } else if (qName.equals("resourcePartition")) {
                if (this.currentResourcePartition != null) {
                    String partitionName = attrs.getValue("name");
                    if (partitionName == null) {
                        DomainBuilder.logger.error("Missing partitionName name, skipping partitionName definition");
                        return;
                    }
                    this.currentPath.append("/" + partitionName);
                    try {
                        long cpuCores = Long.parseLong(attrs.getValue("cpuCores"));
                        long memoryMB = Long.parseLong(attrs.getValue("memoryMB"));
                        long storageMB = Long.parseLong(attrs.getValue("storageMB"));
                        ResourcePartition part = (ResourcePartition) this.currentResourcePartition.createSubResourcePartition(
                            partitionName, cpuCores, memoryMB, storageMB, DomainBuilder.attributes2Map(attrs));
                        this.resourcePartitionStack.push(part);
                        this.currentResourcePartition = part;
                    } catch (VMMException ex) {
                        DomainBuilder.logger.error("Skipping partition " + partitionName, ex);
                    }
                }
            } else if (qName.equals("host")) {
                String hostName = attrs.getValue("name");
                if (hostName == null) {
                    DomainBuilder.logger.error("Missing host name, skipping host definition");
                    return;
                }
                if (this.currentServerPool != null) {
                    try {
                        this.currentServerPool.newHost(hostName, DomainBuilder.attributes2Map(attrs));
                    } catch (VMMException ex) {
                        DomainBuilder.logger.error("Skipping host " + hostName + ":" + ex.getMessage());
                    }
                }
            } else if (qName.equals("storagePool")) {
                String poolName = attrs.getValue("name");
                if (poolName == null) {
                    DomainBuilder.logger.error("Missing storage pool name, skipping pool definition");
                    return;
                }
                if (this.currentServerPool != null) {
                    try {
                        this.currentServerPool.addStoragePool(poolName, DomainBuilder.attributes2Map(attrs));
                    } catch (VMMException ex) {
                        DomainBuilder.logger.error("Skipping storage pool " + poolName + ":" + ex.getMessage());
                    }
                }
            }
        }

        @Override
        public void endElement(final java.lang.String uri, final java.lang.String localName, final java.lang.String qName)
            throws SAXException {
            if (qName.equals("domain")) {
                if (this.currentDomain != null) {
                    this.domainStack.pop();
                    this.currentPath.delete(this.currentPath.lastIndexOf("/"), this.currentPath.length());
                }
                if (this.domainStack.isEmpty()) {
                    this.currentDomain = null;
                } else {
                    this.currentDomain = this.domainStack.peek();
                }
            } else if (qName.equals("serverPool")) {
                if (this.currentServerPool != null) {
                    this.currentServerPool.postCreateHook();
                    this.currentServerPool = null;
                }
                this.resourcePartitionStack.clear();
            } else if (qName.equals("monitoring")) {
                if (this.currentMonitorable != null) {
                    this.currentMonitorable = null;
                }
            } else if (qName.equals("resourcePartition")) {
                if (this.currentResourcePartition != null) {
                    this.resourcePartitionStack.pop();
                    this.currentPath.delete(this.currentPath.lastIndexOf("/"), this.currentPath.length());
                }
                if (this.resourcePartitionStack.isEmpty()) {
                    this.currentResourcePartition = null;
                } else {
                    this.currentResourcePartition = this.resourcePartitionStack.peek();
                }
            }
        }

    }

    static private Map<String, String> attributes2Map(final Attributes attrs) {
        Map<String, String> map = new HashMap<String, String>();
        for (int i = 0; i < attrs.getLength(); i++) {
            String name = attrs.getQName(i);
            String val = attrs.getValue(i);
            map.put(name, val);
        }
        return map;
    }

    /**
     * Constructs a new Domain object from an XML description
     * 
     * @param file the file containing of the XML description
     * @return the domain object or null if the domain could not be created.
     */
    public static Domain createManagedResourcesFromXML(final File file) {
        try {
            DomainXMLHandler handler = new DomainXMLHandler();
            SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
            DomainBuilder.logger.info("Reading resource description file " + file.getAbsolutePath());
            parser.parse(new FileInputStream(file), handler);
            return handler.rootDomain;
        } catch (Exception ex) {
            DomainBuilder.logger.error("While parsing resource description file:", ex);
            return null;
        }
    }

}
