/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: ResourcePartition.java 156 2011-07-19 16:36:14Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.domain;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.agent.main.VirtManagerAgent;
import org.ow2.sirocco.vmm.api.CustomizationSpec;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.IllegalOperationException;
import org.ow2.sirocco.vmm.api.InsufficientResourcesException;
import org.ow2.sirocco.vmm.api.InvalidVMConfigException;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.ResourcePartitionMXBean;
import org.ow2.sirocco.vmm.api.ServerPoolMXBean;
import org.ow2.sirocco.vmm.api.VMConfigSpec;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VirtualMachineConfigSpec;
import org.ow2.sirocco.vmm.api.VirtualMachineImageMXBean;
import org.ow2.sirocco.vmm.api.VirtualMachineMXBean;

public class ResourcePartition extends ManagedResource implements ResourcePartitionMXBean {
    static protected Logger logger = Logger.getLogger(ResourcePartition.class);

    public static int DEFAULT_CPU_OVER_PROVISIONING_RATIO = 5;

    protected ServerPool rootServerPool;

    private ResourcePartition parent;

    private List<ResourcePartition> subPartitions = new CopyOnWriteArrayList<ResourcePartition>();

    private String name;

    private Map<String, String> properties = new HashMap<String, String>();

    protected long cpuCoreCapacity;

    protected long memoryCapacityMB;

    protected long storageCapacityMB;

    private long usedCpuCoreCapacity;

    private long usedMemoryCapacityMB;

    private long usedStorageCapacityMB;

    private int cpuOverprovisioningRatio = ResourcePartition.DEFAULT_CPU_OVER_PROVISIONING_RATIO;

    private List<VirtualMachineMXBean> virtualMachines = new CopyOnWriteArrayList<VirtualMachineMXBean>();

    public ResourcePartition(final ObjectName objName, final String symbolicName, final ResourcePartition parent,
        final Map<String, String> properties) {
        super(objName);
        this.name = symbolicName;
        this.parent = parent;
        if (properties != null && properties.get("cpuOverprovisioningRatio") != null) {
            try {
                this.setCpuOverprovisioningRatio(Integer.parseInt(properties.get("cpuOverprovisioningRatio")), true);
            } catch (InsufficientResourcesException e) {
                // should never occur
            }
        }
        this.properties = new HashMap<String, String>(properties);
    }

    public void delete(final boolean unregister) {
        for (ResourcePartition subPart : this.subPartitions) {
            subPart.delete(true);
        }
        if (unregister) {
            try {
                AgentCommon.getMBeanServer().unregisterMBean(this.objectName);
            } catch (Exception ex) {
                ResourcePartition.logger.error("Failed to unregister ResourcePartition " + this.getName(), ex);
            }
        }
        this.subPartitions.clear();
        this.virtualMachines.clear();
    }

    protected ResourcePartition getSubPartitionByObjectName(final ObjectName objectName) {
        for (ResourcePartition rp : this.subPartitions) {
            if (rp.getObjectName().equals(objectName)) {
                return rp;
            }
            ResourcePartition rp2 = rp.getSubPartitionByObjectName(objectName);
            if (rp2 != null) {
                return rp2;
            }
        }
        return null;
    }

    public ResourcePartition getSubPartitionByPath(final String path) {
        for (ResourcePartition rp : this.subPartitions) {
            if (rp.getPath().equals(path)) {
                return rp;
            }
            ResourcePartition rp2 = rp.getSubPartitionByPath(path);
            if (rp2 != null) {
                return rp2;
            }
        }
        return null;
    }

    private void updateUsedCapacity(final VirtualMachineMXBean vm, final int op) {
        try {
            this.usedCpuCoreCapacity += op * vm.getNumVCPUs();
            this.usedMemoryCapacityMB += op * vm.getMemorySizeMB();
            // XXX disk size not known
        } catch (VMMException ex) {
            ResourcePartition.logger.error("Cannot update used capacity", ex);
        }
    }

    private void updateUsedCapacity(final ResourcePartition child, final int op) {
        this.usedCpuCoreCapacity += op * child.cpuCoreCapacity;
        this.usedMemoryCapacityMB += op * child.memoryCapacityMB;
        this.usedStorageCapacityMB += op * child.storageCapacityMB;
    }

    private void updateUsedCapacity(final long deltaCpu, final long deltaMemory, final long deltaStorage) {
        this.usedCpuCoreCapacity += deltaCpu;
        this.usedMemoryCapacityMB += deltaMemory;
        this.usedStorageCapacityMB += deltaStorage;
    }

    @Override
    public String getName() {
        return this.name;
    }

    @Override
    public Map<String, String> getProperties() {
        return new HashMap<String, String>(this.properties);
    }

    @Override
    public int getCpuOverprovisioningRatio() {
        return this.cpuOverprovisioningRatio;
    }

    @Override
    public void setCpuOverprovisioningRatio(final int ratio, final boolean propagate) throws InsufficientResourcesException {
        // XXX check free resources
        this.cpuOverprovisioningRatio = ratio;
        if (propagate) {
            for (ResourcePartition subPart : this.subPartitions) {
                subPart.setCpuOverprovisioningRatio(ratio, true);
            }
        }
        VirtManagerAgent.getInstance().setResourceTreeDirty();
        this.emitNotification(NotificationTypes.RESOURCE_PARTITION_INVENTORY_UPDATE,
            "Update Resource Partition " + this.getPath(), this.getObjectName());
    }

    @Override
    public long getCpuCoreCapacity() {
        return this.cpuCoreCapacity;
    }

    @Override
    public long getMemoryCapacityMB() {
        return this.memoryCapacityMB;
    }

    @Override
    public long getStorageCapacityMB() {
        return this.storageCapacityMB;
    }

    @Override
    public long getFreeCpuCoreCapacity() {
        return this.cpuCoreCapacity * this.cpuOverprovisioningRatio - this.usedCpuCoreCapacity;
    }

    @Override
    public long getFreeMemoryCapacityMB() {
        return this.memoryCapacityMB - this.usedMemoryCapacityMB;
    }

    @Override
    public long getFreeStorageMBCapacity() {
        return this.storageCapacityMB - this.usedStorageCapacityMB;
    }

    public void addVirtualMachine(final VirtualMachineMXBean vm) {
        if (this.virtualMachines.add(vm)) {
            this.updateUsedCapacity(vm, 1);
        }
    }

    public void removeVirtualMachine(final VirtualMachineMXBean vm) {
        if (this.virtualMachines.remove(vm)) {
            this.updateUsedCapacity(vm, -1);
        }
    }

    @Override
    public VirtualMachineMXBean provisionVM(final VMConfigSpec vmSpec, final Map<String, String> constraints, final boolean sync)
        throws InsufficientResourcesException, InvalidVMConfigException, VMMException {
        VirtualMachineImageMXBean image = this.rootServerPool.getVMImageStore().lookUpByUUID(vmSpec.getVmImageUUID());
        if (image == null) {
            throw new InvalidVMConfigException("Illegal VMImage UUID " + vmSpec.getVmImageUUID());
        }
        if (vmSpec.getNumVCPU() > this.getFreeCpuCoreCapacity() || vmSpec.getMemorySizeMB() > this.getFreeMemoryCapacityMB()
            || vmSpec.getDiskSizeMB() > this.getFreeStorageMBCapacity()) {
            ResourcePartition.logger.debug("Failed to provision VM " + vmSpec.getName() + ": not enough resource");
            ResourcePartition.logger
                .debug("\tCPU:  required=" + vmSpec.getNumVCPU() + " free=" + this.getFreeCpuCoreCapacity());
            ResourcePartition.logger.debug("\tRAM:  required=" + vmSpec.getMemorySizeMB() + "MB free="
                + this.getFreeMemoryCapacityMB() + "MB");
            ResourcePartition.logger.debug("\tDISK:  required=" + vmSpec.getDiskSizeMB() + "MB free="
                + this.getFreeStorageMBCapacity() + "MB");
            throw new InsufficientResourcesException();
        }
        HostMXBean candidateHost = this.rootServerPool.getPlacementPolicy().placeVM(this.rootServerPool, vmSpec);
        if (candidateHost == null) {
            throw new InsufficientResourcesException();
        }
        return candidateHost.createVM(vmSpec, this, sync);
    }

    @Override
    public VirtualMachineMXBean createVirtualMachine(final VirtualMachineConfigSpec vmConfigSpec,
        final CustomizationSpec customizationSpec, final boolean startVM, final boolean sync)
        throws InsufficientResourcesException, InvalidVMConfigException, VMMException {
        // if (vmConfigSpec.getNumVCPUs() > this.getFreeCpuCoreCapacity()
        // || vmConfigSpec.getMemoryMB() > this.getFreeMemoryCapacityMB()
        // || vmConfigSpec.computeDiskSpaceRequirementMB() >
        // this.getFreeStorageMBCapacity()) {
        // ResourcePartition.logger.debug("Failed to provision VM " +
        // vmConfigSpec.getName() + ": not enough resource");
        // ResourcePartition.logger.debug("\tCPU:  required=" +
        // vmConfigSpec.getNumVCPUs() + " free="
        // + this.getFreeCpuCoreCapacity());
        // ResourcePartition.logger.debug("\tRAM:  required=" +
        // vmConfigSpec.getMemoryMB() + "MB free="
        // + this.getFreeMemoryCapacityMB() + "MB");
        // ResourcePartition.logger.debug("\tDISK:  required=" +
        // vmConfigSpec.computeDiskSpaceRequirementMB() + "MB free="
        // + this.getFreeStorageMBCapacity() + "MB");
        // throw new InsufficientResourcesException();
        // }
        HostMXBean candidateHost = this.rootServerPool.getPlacementPolicy().placeVM(this.rootServerPool, vmConfigSpec);
        if (candidateHost == null) {
            throw new InsufficientResourcesException();
        }
        vmConfigSpec.setPartition(this);
        return candidateHost.createVirtualMachine(vmConfigSpec, customizationSpec, startVM, sync);
    }

    @Override
    public List<VirtualMachineMXBean> getVirtualMachines() throws VMMException {
        return this.virtualMachines;
    }

    @Override
    public void moveIntoResourcePartition(final VirtualMachineMXBean vm) throws IllegalOperationException, VMMException {
        if (vm.getResourcePartition() == null) {
            throw new IllegalOperationException("VM " + vm.getNameLabel() + " nto attached to a resource pool");
        }
        ResourcePartition fromRP = this.rootServerPool.getSubPartitionByObjectName(vm.getResourcePartition().getObjectName());
        if (fromRP == null) {
            throw new IllegalOperationException("Cannot move VM to a different ServerPool");
        }
        // XXX storage capacity not check
        if (vm.getNumVCPUs() > this.getFreeCpuCoreCapacity() || vm.getMemorySizeMB() > this.getFreeMemoryCapacityMB()) {
            throw new InsufficientResourcesException();
        }
        ((AbstractVirtualMachine) vm).setResourcePartition(this);
    }

    @Override
    public synchronized void changeCapacity(final long newCpuCoreCapacity, final long newMemoryCapacityMB,
        final long newStorageCapacityMB) throws InsufficientResourcesException, IllegalOperationException {
        if (this.parent == null) {
            throw new IllegalOperationException("Cannot change the capacity of a serverpool");
        }
        long deltaCpu = newCpuCoreCapacity - this.cpuCoreCapacity;
        long deltaMemory = newMemoryCapacityMB - this.memoryCapacityMB;
        long deltaStorage = newStorageCapacityMB - this.storageCapacityMB;
        if ((deltaCpu > 0 && this.parent.getFreeCpuCoreCapacity() < deltaCpu * this.parent.cpuOverprovisioningRatio)
            || (deltaMemory > 0 && this.parent.getFreeMemoryCapacityMB() < deltaMemory)
            || (deltaStorage > 0 && this.parent.getFreeStorageMBCapacity() < deltaStorage)) {
            throw new InsufficientResourcesException("Not enough free capacity in the parent partition");
        }
        if ((deltaCpu < 0 && this.getFreeCpuCoreCapacity() < -deltaCpu * this.cpuOverprovisioningRatio)
            || (deltaMemory < 0 && this.getFreeMemoryCapacityMB() < -deltaMemory)
            || (deltaStorage < 0 && this.getFreeStorageMBCapacity() < deltaStorage)) {
            throw new IllegalOperationException("New capacity cannot accomodate partition content");
        }
        this.cpuCoreCapacity = newCpuCoreCapacity;
        this.memoryCapacityMB = newMemoryCapacityMB;
        this.storageCapacityMB = newStorageCapacityMB;
        this.parent.updateUsedCapacity(deltaCpu, deltaMemory, deltaStorage);
        VirtManagerAgent.getInstance().setResourceTreeDirty();
        this.emitNotification(NotificationTypes.RESOURCE_PARTITION_INVENTORY_UPDATE,
            "Update Resource Partition " + this.getPath(), this.getObjectName());
    }

    @Override
    public synchronized ResourcePartitionMXBean createSubResourcePartition(final String name, final long cpuCoreCapacity,
        final long memoryCapacityMB, final long storageCapacityMB, final Map<String, String> properties)
        throws InsufficientResourcesException, VMMException {
        for (ResourcePartition rp : this.subPartitions) {
            if (rp.getName().equals(name)) {
                throw new VMMException("Name " + name + " already used");
            }
        }
        ResourcePartition.logger.debug("freeCpu=" + this.getFreeCpuCoreCapacity() + " freeMem="
            + this.getFreeMemoryCapacityMB() + " storage=" + this.getFreeStorageMBCapacity());
        // if (cpuCoreCapacity > this.getFreeCpuCoreCapacity() ||
        // memoryCapacityMB > this.getFreeMemoryCapacityMB()
        // || storageCapacityMB > this.getFreeStorageMBCapacity()) {
        // throw new InsufficientResourcesException();
        // }

        ObjectName subPartitionObjectName = MBeanObjectNamer.makeResourcePartitionName(this.getPath() + "/" + name);
        ResourcePartition subPartition = new ResourcePartition(subPartitionObjectName, name, this, properties);
        subPartition.rootServerPool = this.rootServerPool;
        subPartition.cpuCoreCapacity = cpuCoreCapacity;
        subPartition.memoryCapacityMB = memoryCapacityMB;
        subPartition.storageCapacityMB = storageCapacityMB;

        if (properties != null) {
            if (properties.get("cpuOverprovisioningRatio") != null) {
                subPartition.setCpuOverprovisioningRatio(Integer.parseInt(properties.get("cpuOverprovisioningRatio")), true);
            } else {
                subPartition.setCpuOverprovisioningRatio(this.cpuOverprovisioningRatio, true);
            }
        }

        try {
            AgentCommon.getMBeanServer().registerMBean(subPartition, subPartitionObjectName);
        } catch (Exception ex) {
            ResourcePartition.logger.error("Failed to register ResourcePartition MXBean", ex);
            throw new VMMException("Internal error", ex);
        }
        this.subPartitions.add(subPartition);
        this.updateUsedCapacity(subPartition, 1);
        ResourcePartition.logger.info("Added resourcePartition " + subPartition.getPath() + " cpuCores="
            + subPartition.cpuCoreCapacity + " memoryMB=" + subPartition.memoryCapacityMB + " storageMB="
            + subPartition.storageCapacityMB + " cpuOverprovisioningRatio=" + subPartition.cpuOverprovisioningRatio);
        VirtManagerAgent.getInstance().setResourceTreeDirty();
        this.emitNotification(NotificationTypes.RESOURCE_PARTITION_INVENTORY_ADD,
            "New Resource Partition " + subPartition.getPath(), subPartition.getObjectName());
        return subPartition;
    }

    @Override
    public List<ResourcePartitionMXBean> getSubResourcePartitions() {
        return new ArrayList<ResourcePartitionMXBean>(this.subPartitions);
    }

    @Override
    public ResourcePartitionMXBean getParent() {
        return this.parent;
    }

    @Override
    public ServerPoolMXBean getRootServerPool() {
        return this.rootServerPool;
    }

    @Override
    public synchronized void removeSubResourcePartition(final ResourcePartitionMXBean partition) throws VMMException {
        for (ResourcePartition subPart : this.subPartitions) {
            if (subPart.getPath().equals(partition.getPath())) {
                for (ResourcePartition subsubPart : subPart.subPartitions) {
                    subPart.removeSubResourcePartition(subsubPart);
                }
                // reparent VMs
                for (VirtualMachineMXBean vmBean : subPart.virtualMachines) {
                    AbstractVirtualMachine vm = (AbstractVirtualMachine) vmBean;
                    vm.setResourcePartition(this);
                }
                subPart.virtualMachines.clear();
                try {
                    AgentCommon.getMBeanServer().unregisterMBean(subPart.getObjectName());
                } catch (Exception ex) {
                    ResourcePartition.logger.error("Failed to unregister ResourcePartition " + this.getName(), ex);
                }
                this.subPartitions.remove(subPart);
                this.updateUsedCapacity(subPart, -1);
                VirtManagerAgent.getInstance().setResourceTreeDirty();
                this.emitNotification(NotificationTypes.RESOURCE_PARTITION_INVENTORY_DEL, "Delete Resource Partition "
                    + subPart.getPath(), subPart.getObjectName());
                return;
            }
        }
        throw new VMMException("Partition " + partition.getPath() + " not a child of partition " + this.getPath());
    }

}
