/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: ServerPool.java 325 2011-09-15 11:57:02Z chou $
 *
 */

package org.ow2.sirocco.vmm.agent.domain;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.ObjectName;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.policy.PolicyRegistry;
import org.ow2.sirocco.vmm.agent.domain.policy.VMPlacementPolicy;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.AgentCommon;
import org.ow2.sirocco.vmm.agent.main.VirtManagerAgent;
import org.ow2.sirocco.vmm.agent.monitoring.api.MonitoringDriverAPI;
import org.ow2.sirocco.vmm.api.HostMXBean;
import org.ow2.sirocco.vmm.api.NotificationTypes;
import org.ow2.sirocco.vmm.api.ServerPoolMXBean;
import org.ow2.sirocco.vmm.api.StoragePoolMXBean;
import org.ow2.sirocco.vmm.api.VMMException;
import org.ow2.sirocco.vmm.api.VirtualMachineImageStoreMXBean;
import org.ow2.sirocco.vmm.api.monitoring.PerfMetric;
import org.ow2.sirocco.vmm.api.monitoring.PerfMetricInfo;

/**
 * Base class partially implementing the ServerPoolMXBean interface and expected
 * to be sub-classed on a per driver basis.
 * <p>
 * A ServerPool is associated with a VM placement policy which determines where
 * a provisioned VM is to be placed on one host among the members of the server
 * pool
 */
public abstract class ServerPool extends ResourcePartition implements ServerPoolMXBean {
    static protected Logger logger = Logger.getLogger(ServerPool.class);

    protected String serverPoolName;

    private MonitoringDriverAPI monitoring;

    protected Map<String, String> attributes = new HashMap<String, String>();

    private final List<AbstractHost> hosts = new ArrayList<AbstractHost>();

    private List<AbstractStoragePool> storagePools = new ArrayList<AbstractStoragePool>();

    protected VMPlacementPolicy placementPolicy = PolicyRegistry.getInstance().newDefaultPolicy();

    /**
     * Constructs a new server pool
     * 
     * @param name symbolic name of the server pool
     * @param objectName JMX object name of the server pool
     */
    public ServerPool(final String name, final ObjectName objectName, final Map<String, String> attributes) {
        super(objectName, "default_pool", null, attributes);
        this.rootServerPool = this;
        this.serverPoolName = name;
        this.attributes = attributes;
    }

    /*
     * @see org.ow2.sirocco.vmm.api.ServerPoolMXBean#getName()
     */
    @Override
    public String getName() {
        return this.serverPoolName;
    }

    abstract public HostMXBean newHost(String hostName, Map<String, String> props) throws VMMException;

    public Map<String, String> getAttributes() {
        return this.attributes;
    }

    public VMPlacementPolicy getPlacementPolicy() {
        return this.placementPolicy;
    }

    /**
     * Adds a new host to the server pool
     * 
     * @param host host to add
     */
    protected void addHost(final AbstractHost host) {
        this.hosts.add(host);
        try {
            // Monitoring
            if (this.monitoring != null) {
                this.monitoring.startHost(host.getHostName());
            }
            this.cpuCoreCapacity += host.getNumCPU();
            this.memoryCapacityMB += host.getTotalMemoryMB();
            ServerPool.logger.info("Added host " + host.getHostName());
        } catch (VMMException ex) {
            ServerPool.logger.error(ex);
        }
    }

    /*
     * @see org.ow2.sirocco.vmm.api.ServerPoolMXBean#addHost(java.util.Map)
     */
    public HostMXBean addHost(final Map<String, String> attributes) throws VMMException {
        String hostName = attributes.get("name");
        if (hostName == null) {
            throw new VMMException("Missing host name attribute");
        }
        for (HostMXBean host : this.hosts) {
            if (host.getHostName().equals(hostName)) {
                throw new VMMException("Host " + hostName + " already exists");
            }
        }
        HostMXBean host = null;
        try {
            host = this.newHost(hostName, attributes);
        } catch (IllegalArgumentException ex) {
            DomainBuilder.logger.error("Cannot add host " + hostName + ":" + ex.getMessage());
            throw new VMMException(ex.getMessage());
        }
        if (host != null) {
            this.emitNotification(NotificationTypes.HOST_INVENTORY_ADD, "Ajout host", host.getObjectName());
            VirtManagerAgent.getInstance().setResourceTreeDirty();
        }
        return host;
    }

    protected void release() {
    }

    public synchronized void delete() {
        // TODO stop monitoring
        for (AbstractHost host : this.hosts) {
            host.delete();
        }
        this.hosts.clear();

        // resource partition
        super.delete(false);

        this.deleteImageStore();

        for (AbstractStoragePool storagePool : this.storagePools) {
            storagePool.delete();
        }
        this.storagePools.clear();

        try {
            AgentCommon.getMBeanServer().unregisterMBean(this.objectName);
        } catch (Exception ex) {
            ServerPool.logger.error(ex);
        }
        this.release();
        VirtManagerAgent.getInstance().setResourceTreeDirty();
        this.emitNotification(NotificationTypes.SERVER_POOL_INVENTORY_DEL, "delete serverpool", this.getObjectName());
    }

    /*
     * @see org.ow2.sirocco.vmm.api.ServerPoolMXBean#deleteHost(org.ow2.sirocco.
     * vmm.api.HostMXBean)
     */
    public void deleteHost(final HostMXBean host) throws VMMException {
        ObjectName hostObjectName = null;

        hostObjectName = host.getObjectName();

        int indexToRemove = -1;
        for (HostMXBean h : this.hosts) {
            if (h.getObjectName().equals(hostObjectName)) {
                indexToRemove = this.hosts.indexOf(h);
            }
        }
        if (indexToRemove == -1) {
            throw new VMMException("Cannot find host " + host.getHostName());
        }
        AbstractHost hostToRemove = this.hosts.get(indexToRemove);
        this.cpuCoreCapacity -= host.getNumCPU();
        this.memoryCapacityMB -= host.getTotalMemoryMB();
        // Monitoring
        if (this.monitoring != null) {
            this.monitoring.stopHost(host.getHostName());
        }
        this.hosts.remove(indexToRemove);
        hostToRemove.delete();
    }

    /*
     * @see
     * org.ow2.sirocco.vmm.api.ServerPoolMXBean#moveHost(org.ow2.sirocco.vmm
     * .api.HostMXBean, org.ow2.sirocco.vmm.api.ServerPoolMXBean)
     */
    public void moveHost(final HostMXBean host, final ServerPoolMXBean server) throws VMMException {
        Map<String, String> attributes = host.getAttributes();
        this.deleteHost(host);
        server.addHost(attributes);
    }

    public void setProvisioningPolicy(final String policyId) {
        VMPlacementPolicy newPlacementPolicy = PolicyRegistry.getInstance().newPolicy(policyId);
        if (newPlacementPolicy != null) {
            this.placementPolicy = newPlacementPolicy;
            ServerPool.logger.info("Domain " + this.serverPoolName + ": policy set to " + policyId);
        }
    }

    public List<StoragePoolMXBean> getStoragePools() {
        return new ArrayList<StoragePoolMXBean>(this.storagePools);
    }

    protected abstract AbstractStoragePool newStoragePool(String name, ObjectName objectName, Map<String, String> attributes)
        throws Exception;

    @Override
    public void addStoragePool(final String name, final Map<String, String> attributes) throws VMMException {
        synchronized (this.storagePools) {
            for (AbstractStoragePool storagePool : this.storagePools) {
                if (storagePool.getName().equals(name)) {
                    throw new VMMException("Duplicated storage pool name " + name);
                }
            }
            ObjectName storagePoolObjectName = MBeanObjectNamer.makeStoragePoolName(this.getPath() + "/" + name);
            AbstractStoragePool pool;
            try {
                pool = this.newStoragePool(name, storagePoolObjectName, attributes);
            } catch (Exception ex) {
                ServerPool.logger.error("Failed to create StoragePool ", ex);
                throw new VMMException("Failed to create storage pool " + ex.getMessage());
            }
            try {
                AgentCommon.getMBeanServer().registerMBean(pool, storagePoolObjectName);
            } catch (Exception ex) {
                ServerPool.logger.error("Failed to register StoragePool MXBean " + storagePoolObjectName);
            }
            this.storagePools.add(pool);
            ServerPool.logger.info("Added storage pool " + name);
        }
    }

    @Override
    public void removeStoragePool(final StoragePoolMXBean storagePool) throws VMMException {
        synchronized (this.storagePools) {
            for (StoragePoolMXBean pool : this.storagePools) {
                if (pool.getObjectName().equals(storagePool.getObjectName())) {
                    this.storagePools.remove(pool);
                    ((AbstractStoragePool) pool).delete();
                }
            }
        }
        throw new VMMException("Storage pool " + storagePool.getObjectName() + " not a member of server pool "
            + this.serverPoolName);
    }

    // public StoragePoolMXBean getStoragePoolByProperty(final String property)
    // {
    // for (StoragePoolMXBean storagePool : this.getStoragePools()) {
    // if (storagePool.getProperties().get(property) != null) {
    // return storagePool;
    // }
    // }
    // return null;
    // }
    //
    public StoragePoolMXBean getStoragePoolByName(final String name) {
        for (StoragePoolMXBean storagePool : this.getStoragePools()) {
            if (storagePool.getName().equals(name)) {
                return storagePool;
            }
        }
        return null;
    }

    /*
     * @see org.ow2.sirocco.vmm.api.ServerPoolMXBean#getVMImageStore()
     */
    abstract public VirtualMachineImageStoreMXBean getVMImageStore();

    abstract public void deleteImageStore();

    /*
     * @see org.ow2.sirocco.vmm.api.ServerPoolMXBean#getManagedHosts()
     */
    public List<HostMXBean> getManagedHosts() {
        return new ArrayList<HostMXBean>(this.hosts);
    }

    /*
     * @see org.ow2.sirocco.vmm.api.ServerPoolMXBean#getLatestHost()
     */
    public HostMXBean getLatestHost() {
        return this.hosts.get(this.hosts.size() - 1);
    }

    public void postCreateHook() {
    }

    @Override
    public MBeanNotificationInfo[] getNotificationInfo() {
        return new MBeanNotificationInfo[] {new MBeanNotificationInfo(new String[] {NotificationTypes.HOST_INVENTORY_ADD,
            NotificationTypes.HOST_INVENTORY_DEL, NotificationTypes.SERVER_POOL_INVENTORY_ADD,
            NotificationTypes.SERVER_POOL_INVENTORY_DEL, NotificationTypes.LOG, NotificationTypes.ERROR},
            Notification.class.getName(), "Server Pool event")};
    }

    public void setMonitoringDriver(final MonitoringDriverAPI monitoring) {
        this.monitoring = monitoring;
    }

    public MonitoringDriverAPI getMonitoringDriver() {
        return this.monitoring;
    }

    public PerfMetricInfo[] listPerfMetricInfos() throws VMMException {
        return this.getMonitoringDriver().listPerfMetrics(this);
    }

    public PerfMetric getPerfMetric(final String info) throws VMMException {
        return this.getMonitoringDriver().getPerfMetric(this, info);

    }

    public PerfMetric[] getPerfMetrics(final String metricName, final Date startTime, final Date endTime, final long period,
        final ConsolidationFunction conFun) throws VMMException {
        return this.monitoring.getPerfMetrics(this, metricName, startTime, endTime, period, conFun.toString());

    }

}
