/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: RemoteExec.java 13 2011-06-14 21:21:51Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.driver.util;

import java.io.InputStream;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.apache.log4j.Logger;

import com.jcraft.jsch.ChannelExec;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.Session;
import com.jcraft.jsch.UserInfo;

/**
 * Utility class allowing to invoke commands on a remote host using SSH
 */
public class RemoteExec {
    static Logger logger = Logger.getLogger(RemoteExec.class);

    /**
     * Result of a remote SSH invocation
     */
    static public class Result {
        /**
         * exit code of the remote command
         */
        public int exitCode;

        /**
         * standard output
         */
        public String output;

        /**
         * standard error
         */
        public String error;
    }

    /**
     * SSH authentification information
     */
    public static class SshAuthInfo {
        private String password;

        private String keyFile;

        /**
         * Constructs a new SshAuthInfo based on either a password or a private
         * key location
         * 
         * @param password password or null if a key file is provided
         * @param keyFile filename of a private SSH key file or null if a
         *        password is provided
         */
        public SshAuthInfo(final String password, final String keyFile) {
            this.password = password;
            this.keyFile = keyFile;
            if (password == null && keyFile == null) {
                throw new IllegalArgumentException();
            }
        }

        @Override
        public String toString() {
            return (this.password != null) ? "password=" + this.password : "keyfile=" + this.keyFile;
        }
    }

    static public class SshException extends Exception {
        private static final long serialVersionUID = -7284013737516587531L;
    }

    /**
     * Invoke a command through SSH on a remote host as root
     * 
     * @param host target host
     * @param password root password
     * @param command command to exec on the remote host as root
     * @return result of the invocation
     * @throws SshException raised if the SSH connection could not be
     *         established
     */
    static public Result commandAsRoot(final String host, final SshAuthInfo authInfo, final String command) throws SshException {
        return RemoteExec.command(host, "root", authInfo, command);
    }

    /**
     * Invoke a command through SSH on a remote host
     * 
     * @param host target host
     * @param user user login
     * @param password user password
     * @param command command to exec on the remote host
     * @return result of the invocation
     * @throws SshException raised if the SSH connection could not be
     *         established
     */
    static public Result command(final String host, final String user, final SshAuthInfo authInfo, final String command)
        throws SshException {
        try {
            JSch jsch = new JSch();
            if (authInfo.keyFile != null) {
                jsch.addIdentity(authInfo.keyFile, "passphrase");
            }
            Session session = jsch.getSession(user, host, 22);
            session.setUserInfo(new UserInfo() {
                public String getPassphrase() {
                    return null;
                }

                public String getPassword() {
                    return authInfo.password;
                }

                public boolean promptPassphrase(final String arg0) {
                    return true;
                }

                public boolean promptPassword(final String arg0) {
                    return true;
                }

                public boolean promptYesNo(final String arg0) {
                    return true;
                }

                public void showMessage(final String arg0) {
                }
            });
            session.connect();
            ChannelExec channel = (ChannelExec) session.openChannel("exec");
            (channel).setCommand(command);
            // channel.setInputStream(null);
            // ((ChannelExec) channel).setErrStream(System.err);
            InputStream in = channel.getInputStream();
            InputStream err = channel.getErrStream();
            channel.connect();
            Result result = new Result();
            StringBuffer output = new StringBuffer();
            StringBuffer error = new StringBuffer();
            byte[] tmp = new byte[1024];

            while (true) {
                while (in.available() > 0) {
                    int i = in.read(tmp, 0, 1024);
                    if (i < 0) {
                        break;
                    }
                    String s = new String(tmp, 0, i);
                    output.append(s);
                }
                if (channel.isClosed()) {
                    result.exitCode = channel.getExitStatus();
                    break;
                }
                try {
                    Thread.sleep(100);
                } catch (Exception ee) {
                }
            }

            while (true) {
                while (err.available() > 0) {
                    int i = err.read(tmp, 0, 1024);
                    if (i < 0) {
                        break;
                    }
                    String s = new String(tmp, 0, i);
                    error.append(s);
                }
                if (channel.isClosed()) {
                    result.exitCode = channel.getExitStatus();
                    break;
                }
                try {
                    Thread.sleep(100);
                } catch (Exception ee) {
                }
            }

            channel.disconnect();
            session.disconnect();
            result.output = new String(output);
            result.error = new String(error);
            return result;
        } catch (Exception ex) {
            RemoteExec.logger.error("SSH failure: user=[" + user + "] auth=[" + authInfo + "]", ex);
            SshException e = new SshException();
            e.initCause(ex);
            throw e;
        }
    }

    private static ExecutorService executorService = Executors.newFixedThreadPool(10);

    private static class MyUserInfo implements UserInfo {
        String password;

        MyUserInfo(final String password) {
            this.password = password;
        }

        public String getPassphrase() {
            return null;
        }

        public String getPassword() {
            return this.password;
        }

        public boolean promptPassphrase(final String arg0) {
            return true;
        }

        public boolean promptPassword(final String arg0) {
            return true;
        }

        public boolean promptYesNo(final String arg0) {
            return true;
        }

        public void showMessage(final String arg0) {
        }
    }

    private static class ReaderTask implements Runnable {
        ChannelExec channel;

        SSHExecCallback callback;

        InputStream in;

        boolean stop = false;

        ReaderTask(final ChannelExec channel, final InputStream in, final SSHExecCallback callback) {
            this.channel = channel;
            this.callback = callback;
            this.in = in;
        }

        synchronized void stop() {
            this.stop = true;
        }

        @Override
        public void run() {
            byte[] tmp = new byte[1024];

            while (true) {
                synchronized (this) {
                    if (this.stop) {
                        break;
                    }
                }
                try {
                    while (this.in.available() > 0) {
                        int i = this.in.read(tmp, 0, 1024);
                        if (i < 0) {
                            break;
                        }
                        String s = new String(tmp, 0, i);
                        this.callback.newData(s);
                    }
                } catch (Exception ex) {
                    break;
                }
                if (this.channel.isClosed()) {
                    break;
                }
                try {
                    Thread.sleep(100);
                } catch (Exception ee) {
                }
            }
        }
    }

    public static interface SSHExecCallback {
        void newData(String s);
    }

    public static class SSHExec {
        private String host;

        private String user;

        private SshAuthInfo authInfo;

        private String command;

        private int exitCode;

        private ChannelExec channel;

        private Session session;

        private ReaderTask outReader, errReader;

        public SSHExec(final String host, final String user, final SshAuthInfo authInfo, final String command) {
            this.host = host;
            this.user = user;
            this.authInfo = authInfo;
            this.command = command;
        }

        public void connect(final SSHExecCallback outCallback, final SSHExecCallback errCallback) throws SshException {
            try {
                JSch jsch = new JSch();
                if (this.authInfo.keyFile != null) {
                    jsch.addIdentity(this.authInfo.keyFile, "passphrase");
                }
                Session session = jsch.getSession(this.user, this.host, 22);
                session.setUserInfo(new MyUserInfo(this.authInfo.password));
                session.connect();
                this.channel = (ChannelExec) session.openChannel("exec");
                (this.channel).setCommand(this.command);
                final InputStream in = this.channel.getInputStream();
                final InputStream err = this.channel.getErrStream();
                this.channel.connect();

                this.outReader = new ReaderTask(this.channel, in, outCallback);
                this.errReader = new ReaderTask(this.channel, err, errCallback);
                RemoteExec.executorService.execute(this.outReader);
                RemoteExec.executorService.execute(this.errReader);

            } catch (Exception ex) {
                RemoteExec.logger.error("SSH failure: user=[" + this.user + "] auth=[" + this.authInfo + "]", ex);
                SshException e = new SshException();
                e.initCause(ex);
                throw e;
            }

        }

        public void disconnect() {
            if (this.outReader != null) {
                this.outReader.stop();
            }
            if (this.errReader != null) {
                this.errReader.stop();
            }
            if (this.channel != null) {
                this.exitCode = this.channel.getExitStatus();
                this.channel.disconnect();
            }
            if (this.session != null) {
                this.session.disconnect();
            }
        }

        public int getExitCode() {
            return this.exitCode;
        }
    }

}
