/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: VirtManagerAgent.java 188 2011-07-25 13:39:39Z dangtran $
 *
 */

package org.ow2.sirocco.vmm.agent.main;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;

import javax.management.ObjectName;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.domain.Domain;
import org.ow2.sirocco.vmm.agent.domain.DomainBuilder;
import org.ow2.sirocco.vmm.agent.driver.Driver;
import org.ow2.sirocco.vmm.agent.driver.DriverFactory;
import org.ow2.sirocco.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.sirocco.vmm.agent.main.persistence.ConfigurationGeneratorException;
import org.ow2.sirocco.vmm.agent.main.persistence.RessourceConfigurationGenerator;
import org.ow2.sirocco.vmm.api.DomainMXBean;
import org.ow2.sirocco.vmm.api.VMMAgentMXBean;
import org.ow2.sirocco.vmm.api.VMMException;

/**
 * VirtManagerAgent implements the VMMAgentMXBean interface and is the entry
 * point of the agent. VirtManagerAgent is a singleton. VirtManagerAgent
 * expects:
 * <ul>
 * <li>the vmm.home property to be set to the directory where the VMM agent
 * distribution is installed
 * <li>the etc/agent.properties file to contains agent properties
 * </ul>
 */
public class VirtManagerAgent extends AgentCommon implements VMMAgentMXBean {
    private static Logger logger = Logger.getLogger(VirtManagerAgent.class);

    static VirtManagerAgent instance;

    private JMXConnectorServer cs;

    private Domain rootDomain;

    private volatile boolean loadingResourceTreeInProgress = false;

    private volatile boolean shuttingDown = false;

    static public VirtManagerAgent getInstance() {
        if (VirtManagerAgent.instance == null) {
            VirtManagerAgent.instance = new VirtManagerAgent();
        }
        return VirtManagerAgent.instance;
    }

    /*
     * @see org.ow2.sirocco.vmm.api.VMMAgentMXBean#getRootDomain()
     */
    public DomainMXBean getRootDomain() {
        return this.rootDomain;
    }

    private VirtManagerAgent() {
    }

    public void start() {
        this.start(9999, "managed-resources.xml", false);
    }

    public void start(final int port, final String managedResourcesFileName, final boolean exitOnError) {
        Thread startThread = new Thread(new Runnable() {
            public void run() {
                try {
                    VirtManagerAgent.this.startAsync(port, managedResourcesFileName);
                } catch (Exception e) {
                    if (exitOnError) {
                        System.exit(1);
                    }
                }
            }
        });
        startThread.start();
    }

    private void startAsync(final int port, final String managedResourcesFileName) throws Exception {
        ObjectName agentName = MBeanObjectNamer.makeAgentName();
        try {
            AgentCommon.getMBeanServer().registerMBean(this, agentName);
        } catch (Exception ex) {
            VirtManagerAgent.logger.fatal("Cannot register MXBean " + agentName, ex);
            throw ex;
        }
        VirtManagerAgent.logger.info("Created VMMAgent MBean " + agentName);

        File managedResourcesFile = new File(AgentControl.vmmHome, "etc" + File.separator + managedResourcesFileName);

        this.loadingResourceTreeInProgress = true;
        this.rootDomain = DomainBuilder.createManagedResourcesFromXML(managedResourcesFile);
        if (this.rootDomain == null) {
            VirtManagerAgent.logger.fatal("Failed to load resource file");
            throw new Exception("Failed to load resource file");
        }
        this.loadingResourceTreeInProgress = false;

        Registry registry = null;
        try {
            registry = LocateRegistry.createRegistry(port);
        } catch (RemoteException ex) {
            VirtManagerAgent.logger.warn("Cannot launch RMI registry on port " + port + ", creating new one");
            try {
                LocateRegistry.createRegistry(port);
            } catch (RemoteException ex2) {
                VirtManagerAgent.logger.fatal("Cannot create RMI registry", ex);
                throw ex2;
            }
        }
        try {
            JMXServiceURL url = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://localhost:" + port + "/server");
            this.cs = JMXConnectorServerFactory.newJMXConnectorServer(url, null, AgentCommon.mbs);
            VirtManagerAgent.logger.info("Starting RMI connector server with URL \"" + url + "\"");
            this.cs.start();
            VirtManagerAgent.logger.info("VMM agent ready");
        } catch (Exception ex) {
            VirtManagerAgent.logger.fatal("Failed to start JMX connector server", ex);
            throw ex;
        }

        synchronized (this) {
            try {
                this.wait();
            } catch (InterruptedException e) {
            }
        }
        this.rootDomain.delete();
        try {
            AgentCommon.getMBeanServer().unregisterMBean(agentName);
        } catch (Exception ex) {
        }

        this.rootDomain = null;

        VirtManagerAgent.logger.info("VMM agent terminated");

        // org.apache.log4j.LogManager.shutdown();
        try {
            this.cs.stop();
        } catch (IOException e) {
            System.out.println(e.getMessage());
        }
        this.cs = null;

        UnicastRemoteObject.unexportObject(registry, true);

        // System.exit(0);
    }

    public synchronized void setResourceTreeDirty() {
        if (!this.loadingResourceTreeInProgress && !this.shuttingDown) {
            this.saveResourceTree();
        }
    }

    public boolean isLoadingInProgess() {
        return this.loadingResourceTreeInProgress;
    }

    private synchronized void saveResourceTree() {
        RessourceConfigurationGenerator newXML = null;
        try {
            newXML = new RessourceConfigurationGenerator(AgentControl.vmmHome + File.separator + "etc");
        } catch (ConfigurationGeneratorException ex) {
            VirtManagerAgent.logger.error(ex);
            return;
        }
        try {
            newXML.generate(this.rootDomain);
            VirtManagerAgent.logger.info("Resource tree saved to disk");
        } catch (ConfigurationGeneratorException ex) {
            VirtManagerAgent.logger.error(ex);
        }
    }

    /*
     * @see org.ow2.sirocco.vmm.api.VMMAgentMXBean#shutdown()
     */
    public void shutdown() {
        this.shuttingDown = true;
        synchronized (this) {
            this.notifyAll();
        }
    }

    public String[] getServerPoolCreationAttributes(final String driverName) throws VMMException {
        Driver driver = DriverFactory.getInstance().newDriver(driverName);
        if (driver == null) {
            throw new VMMException("Unsupported driver " + driverName);
        }
        return driver.getServerPoolCreationAttributes();
    }
}
