/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id: MonitoringDriverFactory.java 132 2011-07-11 08:54:11Z chou $
 *
 */

package org.ow2.sirocco.vmm.agent.monitoring.driver;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.ow2.sirocco.vmm.agent.main.AgentControl;

/**
 * The DriverFactory class is a singleton which acts a factory of Driver
 * objects. Driver are identified with symbolic aliases in domain XML files. The
 * mapping between driver aliases and Java Driver classes is specified in the
 * etc/drivermap.properties file.
 */
public class MonitoringDriverFactory {
    static protected Logger logger = Logger.getLogger(MonitoringDriverFactory.class);

    private static final String DRIVERMAP_FILENAME = "monitoringdrivermap.properties";

    static private MonitoringDriverFactory instance;

    private Map<String, String> map = new HashMap<String, String>();

    /**
     * Returns the single DriverFactory instance
     * 
     * @return
     */
    static public MonitoringDriverFactory getInstance() {
        if (MonitoringDriverFactory.instance == null) {
            MonitoringDriverFactory.instance = new MonitoringDriverFactory();
        }
        return MonitoringDriverFactory.instance;
    }

    private MonitoringDriverFactory() {
        Properties properties = new Properties();
        String agentPropFileName = "etc" + File.separator + MonitoringDriverFactory.DRIVERMAP_FILENAME;
        try {
            properties.load(new FileInputStream(new File(AgentControl.vmmHome, agentPropFileName)));
        } catch (IOException ex) {
            MonitoringDriverFactory.logger.warn("Cannot find " + MonitoringDriverFactory.DRIVERMAP_FILENAME + " file");
        }
        for (Enumeration<Object> it = properties.keys(); it.hasMoreElements();) {
            String driverAlias = (String) it.nextElement();
            String driverClass = properties.getProperty(driverAlias);
            this.map.put(driverAlias, driverClass);
            MonitoringDriverFactory.logger.info("Added alias " + driverAlias + " for " + driverClass);

        }
        // built-in drivers
        if (!this.map.containsKey("xen")) {
            this.map.put("xen", "org.ow2.jasmine.vmmapi.driver.xen.XenDriver");
        }
        if (!this.map.containsKey("vmware-vi")) {
            this.map.put("vmware-vi", "org.ow2.jasmine.vmmapi.driver.vmware.VMwareDriver");
        }
        if (!this.map.containsKey("dummy")) {
            this.map.put("dummy", "org.ow2.jasmine.vmmapi.driver.dummy.DummyDriver");
        }
        if (!this.map.containsKey("collectd")) {
            this.map.put("collectd", "org.ow2.jasmine.vmmapi.monitoring.driver.collectd.CollectdDriver");
        }

    }

    /**
     * Builds a new Driver object
     * 
     * @param driverName alias of the driver to create
     * @return the Driver object
     */
    public MonitoringDriver newDriver(final String driverName) {
        String driverClassName = this.map.get(driverName);
        MonitoringDriver driver = null;
        if (driverClassName != null) {
            try {
                driver = ((MonitoringDriver) Class.forName(driverClassName).newInstance());
            } catch (Exception ex) {
                MonitoringDriverFactory.logger.error("Failed to instance driver " + driverName, ex);
            }
        } else {
            MonitoringDriverFactory.logger.error("Unknown driver " + driverName);
        }
        return driver;
    }

}
